/*
   cromemco-fdc-uart.c
   Copyright 2017 Greg Sydney-Smith
   
   2017-12-02 gss. First version. TUART portions Copyright 2016 Udo Munk.
   2018-09-16 gss. Windows arrow keys map to ^E ^S ^D ^X.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   DEVELOPERS AND CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <stdio.h>
#include <windows.h>
#include <conio.h>
#include "sim.h"
#include "cromemco-fdc-uart.h"
#include "vt100.h"
#include "simfun.h"

void user_int(int), quit_int(int), term_int(int); // in simint.c


// with the CPU:
extern int int_int;             // an interrupt is pending/active
extern int int_data;            // the RST instruction or int vector


static int dbg=0;
static int uart0a_int_mask, uart0a_int, uart0a_int_pending, uart0a_rst7;
static int uart0a_tbe, uart0a_rda, rtc;


int fdc_uart_init(void) {
    rtc= 0;
    uart0a_int= 0xff;
    uart0a_tbe= 1;
    uart0a_rda= 0;
    uart0a_int_mask= 0;
    uart0a_int_pending= 0;
    uart0a_rst7= 0;
    // hardwired to vt100 for now
    return 0;
}

int fdc_uart_config(char *name, char *val) {
    int i;
    char *names[] = { "debug", NULL };

    if (val==NULL) val="0";
    for (i=0; names[i]!=NULL && stricmp(name,names[i])!=0; i++) ;
    switch (i) {
    case 0: dbg= atoi(val); break;
    }
    return 0;
}



// 2 ms pulse
void fdc_uart_tick(void) {
    static int count=0;
    static int chkRDA=0;

    if (count++ >= 512) count=0;
    if ((count %    8) ==   0) {                // per 16 mS
        chkRDA=1; // if (vt100_stat_in()&0x40) uart0a_rda=1; // type < 8 chars per sec
    }
    if ((count %  256) ==   0) {                // per 512 mS
        rtc = 1;                                // set RTC interrupt flag
    }

    // if last interrupt not acknowledged by CPU, no new one yet
    if (int_int) return;

    if (chkRDA)
//        if (vt100_stat_in()&0x40) uart0a_rda=1; // type < 8 chars per sec
        if (kbhit()) uart0a_rda=1; // type < 8 chars per sec

    /* UART 0A transmit buffer empty */
    if (uart0a_tbe == 0) {
        uart0a_tbe = 1;
        if (uart0a_int_mask & 32) {
            uart0a_int = 0xef;
            uart0a_int_pending = 1;
            int_data = 0xef;
            int_int = 1;
            return;
        }
    }

    /* UART 0A receive data available */
    if ((uart0a_rda) && (uart0a_int_mask & 16)) {
        uart0a_int = 0xe7;
        uart0a_int_pending = 1;
        int_data = 0xe7;
        int_int = 1;
        return;
    }


    /* UART 0A no pending interrupt */
    uart0a_int = 0xff;
    uart0a_int_pending = 0;
}

void fdc_uart_rt10ms(void) {
    static int count=0;

    count++;

    // if last interrupt not acknowledged by CPU, no new one yet
    if (int_int) return;

    /* 512ms RTC */
    if (count>=51 && uart0a_rst7) {
        count-=51;
        if (uart0a_int_mask & 128) {
            uart0a_int = 0xff;
            uart0a_int_pending = 1;
            int_data = 0xff;
            int_int = 1;
            return;
        }
    }
}

/*
 *	Interface to iosim.c
 */

BYTE fdc_uart_stat_in(void) {       // IN 00
    BYTE status = 0x04;             // bit2 - connected

    if (uart0a_tbe        ) status |= 0x80; // bit7
    if (uart0a_rda        ) status |= 0x40; // bit6
    if (uart0a_int_pending) status |= 0x20; // bit5
    return status;
}

BYTE fdc_uart_data_in(void) {       // IN 01
    int x;

    uart0a_rda = 0;
    if ((x=vt100_data_in()) == 0 || x==0xe0) { // A Windows function key has been pressed
        x=vt100_data_in();          // 2nd byte of sequence
        if (x==0x5C) quit_int(1);   // SHIFT-F9   = back to simulator
        if (x==0x5D) quit_int(1);   // SHIFT-F10  = back to simulator // changed
        setActivityFlag();
        // 20180916
        if (x==0x48) return 'E'&0x1f; // up    arrow to ^E
        if (x==0x4B) return 'S'&0x1f; // left  arrow to ^S
        if (x==0x4D) return 'D'&0x1f; // right arrow to ^D
        if (x==0x50) return 'X'&0x1f; // down  arrow to ^X
        return ' ';
    }
    setActivityFlag();
    return x;
}

void fdc_uart_data_out(BYTE data) { // OUT 01
    uart0a_tbe = 0;                 // not what i'd expect
    vt100_data_out(data);
}

void fdc_uart_cmd(BYTE data) {      // OUT 02
    uart0a_rst7 = (data & 4) ? 1 : 0;
    if (data & 1) {                 // b0=RESET
        uart0a_rda = 0;
        uart0a_tbe = 1;
        uart0a_int_pending = 0;
     }
}

BYTE fdc_uart_int_addr(void) {      // IN 03
    return (BYTE) uart0a_int;
}

void fdc_uart_int_mask(BYTE data) { // OUT 03
    uart0a_int_mask = data;
}





