/*
 * Z80SIM  -  a Z80-CPU simulator
 *
 * Simulator Copyright (C) 2014-2016 by Udo Munk
 * Win Build Copyright (C) 2017-2018 by Greg Sydney-Smith
 *
 * This module of the simulator contains the I/O simulation
 * for a Cromemco Z-1 system
 *
 * History:
 * 15-DEC-14 first version
 * 20-DEC-14 added 4FDC emulation and machine boots CP/M 2.2
 * 28-DEC-14 second version with 16FDC, CP/M 2.2 boots
 * 01-JAN-15 fixed 16FDC, machine now also boots CDOS 2.58 from 8" and 5.25"
 * 01-JAN-15 fixed frontpanel switch settings, added boot flag to fp switch
 * 12-JAN-15 fdc and tu-art improvements, implemented banked memory
 * 02-FEB-15 modified MMU, implemented timers and interrupts
 * 20-FEB-15 bug fix for release 1.25
 * 10-MAR-15 added support for two parallel port lpt devices on TU-ART
 * 26-MAR-15 added support for two serial port tty devices on TU-ART
 * 29-APR-15 added Cromemco DAZZLER to the machine
 * 06-DEC-16 implemented status display and stepping for all machine cycles
 * 22-DEC-16 moved MMU out to the new memory interface module
 * 15-NOV-17 gss. Changes for Win32.
 * 26-NOV-17 gss. Move some items to fdc_tick()
 * 02-DEC-17 gss. Add fdc uart for Cromix / interrupts
 * 02-JAN-18 gss. Add windows front panel
 * 16-SEP-18 gss. Add simctl cmds 0x22, 0xFB, 0xFC; and default (ret "?").
 */

#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
//#include <poll.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include "sim.h"
#include "simglb.h"
//#include "../iodevices/cromemco-tu-art.h"
#include "../iodevices/cromemco-256kz.h"
#include "../iodevices/cromemco-fdc.h"
#include "../iodevices/cromemco-fdc-uart.h"
#include "../iodevices/vt100.h"
#include "memory.h"
#include <conio.h> // gss

#define TRUE   (1==1)
#define FALSE (!TRUE)


int os_sleep(int mS);
void fdc_uart_rt10ms(void);
int  OS_FindFirst(int len, char *buf);
int  OS_FindNext( int len, char *buf);


/*
 *	Forward declarations
 */
void config_io(void);
static void *timing(void *);
static BYTE io_trap_in(void);
static void io_trap_out(BYTE);


/*
 *	This array contains function pointers for every
 *	input I/O port (0 - 255), to do the required I/O.
 */
static BYTE (*port_in[256]) (void) = { };

/*
 *	This array contains function pointers for every
 *	output I/O port (0 - 255), to do the required I/O.
 */
static void (*port_out[256]) (BYTE) = { };


       int lpt1, lpt2;		/* fds for lpt printer files */


// "Error" Device

/*
 *	I/O input trap function
 *	This function should be added into all unused
 *	entries of the input port array. It can stop the
 *	emulation with an I/O error.
 */
static BYTE io_trap_in(void)
{
	//printf("I/O trap in port %02x\r\n", io_port);
	if (i_flag) {
		cpu_error = IOTRAPIN;
		cpu_state = STOPPED;
	}
	return((BYTE) 0xff);
}

/*
 *	I/O output trap function
 *	This function should be added into all unused
 *	entries of the output port array. It can stop the
 *	emulation with an I/O error.
 */
static void io_trap_out(BYTE data)
{
	data = data; /* to avoid compiler warning */

	//printf("I/O trap out port %02x\r\n", io_port);

	if (i_flag) {
		cpu_error = IOTRAPOUT;
		cpu_state = STOPPED;
	}
}


// "Ignore" device

/* Not Used with this combination of device cards.
static BYTE nop_in(void) {
    return 0x1a;
} */

static void nop_out(BYTE data) {
    UNREFERENCED_PARAMETER(data);
}


/// "Simulator Control" Device

static int  simctl_cmd = 0;
static int  simctl_state = 0;
static int  simctl_ptr=0;
static char simctl_val[80+1];
static FILE *hfp[8];

//    resetPTRCmd,                /*  3 reset the PTR device                                              */
//    attachPTRCmd,               /*  4 attach the PTR device                                             */
//    detachPTRCmd,               /*  5 detach the PTR device                                             */
//    attachPTPCmd,               /* 16 attach PTP to the file with name at beginning of CP/M command line*/
//    detachPTPCmd,               /* 17 detach PTP                                                        */
//    getHostOSPathSeparatorCmd,  /* 28 obtain the file path separator of the OS under which SIMH runs    */
//    getHostFilenamesCmd         /* 29 perform wildcard expansion and obtain list of file names          */

// cmd str
static void simctl_out(BYTE data) {
    int n;     

//printf("simctl out=%02x\n",data);
    // state 0 : command byte
    if (simctl_state==0) { simctl_cmd=data; simctl_state=1; simctl_ptr=0; return; }

    // state 1 : arg(s) (eg "n:path\file", "a,b,c" or ""), terminated by \0
    simctl_val[simctl_ptr]= data;
    if (simctl_ptr<80) simctl_ptr++;  // 2018-09-16. too long overwrite 'til \0
    if (data!='\0') return;
    
//printf("simctl cmd=%02x, arg=%s ",simctl_cmd,simctl_val);
    // state 2 : arg(s) complete. act on it.
    switch (simctl_cmd) {
    case 0x01: config_io();
               simctl_val[0]='\0'; break;
    case 0x10: n=atoi(simctl_val);
               hfp[n]=fopen(simctl_val+2,"rb");
               strcpy(simctl_val,(hfp[n]==NULL)?"N":"Y"); break;
    case 0x11: n=atoi(simctl_val);
               hfp[n]=fopen(simctl_val+2,"wb");
               strcpy(simctl_val,(hfp[n]==NULL)?"N":"Y"); break;
    case 0x12: n=atoi(simctl_val);
               hfp[n]=fopen(simctl_val+2,"ab");
               strcpy(simctl_val,(hfp[n]==NULL)?"N":"Y"); break;
    case 0x18: n=atoi(simctl_val); fclose(hfp[n]); hfp[n]=NULL;
               simctl_val[0]='\0'; break;
    case 0x19: n=atoi(simctl_val); rewind(hfp[n]);
               simctl_val[0]='\0'; break;
    case 0x20: OS_FindFirst(80,simctl_val); break;
    case 0x21: OS_FindNext( 80,simctl_val); break;
    case 0x22: sprintf(simctl_val,"%c",OS_PathSep()); break;
    case 0xFB: strcpy(simctl_val,"Cromemco hardware: 16FDC, 256KZ, ZPU"); break;
    case 0xFC: sprintf(simctl_val,"z80sim %s / %s %s",RELEASE,USR_COM,USR_REL); break;
    case 0xFD: OS_getcwd(simctl_val,80); break;
    case 0xFE: exit(0); break;                  // to host OS
    case 0xFF: cpu_state = STOPPED;             // to monitor
               simctl_val[0]='\0'; break;
    default  : strcpy(simctl_val,"?");
    }
//printf("result=[%s]\n",simctl_val);
    simctl_state=0;
    simctl_ptr=0;
}

BYTE simctl_in(void) {
    int result;

    if (simctl_ptr >= 80) return 0;
    result= simctl_val[simctl_ptr];
    if (result != 0) simctl_ptr++;
    return (BYTE) result;
}

// 0=lst, 1=rdr, 2=pun, ...

// PaperTape device (RDR,PUN)
// port_in[ 0x10] 40H=rda 80H=tbe, see Cromemco CPM BIOS
static int pt_rda=0;
static int pt_dat;
BYTE pt_stat(void) {
    int stat;

    stat=0;
    if (!pt_rda && hfp[1]!=NULL && !feof(hfp[1]))
        if ((pt_dat=fgetc(hfp[1]))!=EOF)
            pt_rda=1;
    if (pt_rda      ) stat|=0x40;
    if (hfp[2]!=NULL) stat|=0x80;
//printf("pt_stat=%02x\n",stat);
    return (BYTE) stat;
}
void pt_baud(BYTE data) { data=data; }     // set baudrate. ignored.
BYTE pt_rd(void) {
    if ((pt_stat()&0x40) == 0) return (BYTE) 0x1a;
//printf("pt_rd\n");
    pt_rda=0;
    return (BYTE) pt_dat;
}
void pt_wr(BYTE data) {
    if ((pt_stat()&0x80) == 0) return;     // don't wait, won't happen
    fputc(data,hfp[2]);
}

BYTE lst_tbe(void) {
    return (BYTE)((hfp[0]==NULL)? FALSE : TRUE);
}
void lst_wr(BYTE data) {
    if (!lst_tbe()) return;     // don't wait, won't happen
    fputc(data,hfp[0]);
}



// look in config.txt for lines like "cardname.settingname=value"
void config_io(void) {
    int i, len;
    FILE *fp;
    char line[132+1], *p;
    char *cards[] = { "16fdc", "vt100", NULL };
    int (*funcs[])(char *name, char *val) = {
        cromemco_fdc_config, vt100_config, NULL };

    if ((fp=fopen("config.txt","r")) == NULL) return;

//printf("Reading config.txt ...\n");
    while (fgets(line,132,fp) != NULL) {
        for (i=0; cards[i]!=NULL; i++) {
            len= strlen(cards[i]);
            if (strnicmp(cards[i],line,len)==0 && line[len]=='.') {
                   if ((p=strchr(line+len+1,'=')) != NULL) {*p='\0'; p++; }
                   funcs[i](line+len+1,p);
            }
        }
    }
    fclose(fp);

    cromemco_256kz_init();
    cromemco_fdc_init();
    fdc_uart_init();
    vt100_init();
}
/*
 *	This function is to initiate the I/O devices.
 *	It will be called from the CPU simulation before
 *	any operation with the CPU is possible.
 */
void init_io(void) {
    int i;
    pthread_t thread;

    // defaults
    for (i = 0; i < 256; i++) {
        port_in[ i] = io_trap_in;
        port_out[i] = io_trap_out;
    }

    // plug in the Cromemco cards
    port_in[ 0x00]= fdc_uart_stat_in;
    port_in[ 0x01]= fdc_uart_data_in;
    port_in[ 0x03]= fdc_uart_int_addr;
    port_in[ 0x04]= cromemco_fdc_aux_in;
    port_in[ 0x10]= pt_stat;    // Cromemco CPM 40H=rda 80H=tbe
    port_in[ 0x11]= pt_rd;      // Cromemco CPM
    port_in[ 0x30]= cromemco_fdc_status_in;
    port_in[ 0x31]= cromemco_fdc_track_in;
    port_in[ 0x32]= cromemco_fdc_sector_in;
    port_in[ 0x33]= cromemco_fdc_data_in;
    port_in[ 0x34]= cromemco_fdc_diskflags_in;
    port_in[ 0x40]= cromemco_256kz_in;
//    port_in[ 0x54]= lst_stat;    // Cromemco CPM 20H=busy
    port_in[ 0xFE]= simctl_in;

    port_out[0x00]= nop_out;    // baudrate
    port_out[0x01]= fdc_uart_data_out;
    port_out[0x02]= fdc_uart_cmd;    // HIGHBAUD, INTA Enable, RST7
    port_out[0x03]= fdc_uart_int_mask;
    port_out[0x04]= cromemco_fdc_aux_out;
    port_out[0x10]= pt_baud;    // Cromemco CPM
    port_out[0x11]= pt_wr;      // Cromemco CPM
    port_out[0x30]= cromemco_fdc_cmd_out;
    port_out[0x31]= cromemco_fdc_track_out;
    port_out[0x32]= cromemco_fdc_sector_out;
    port_out[0x33]= cromemco_fdc_data_out;
    port_out[0x34]= cromemco_fdc_diskctl_out;
    port_out[0x40]= cromemco_256kz_out;
//    port_out[0x54]= lst_data;    // Cromemco CPM
    port_out[0xFE]= simctl_out;

    fp_start();
    config_io();

    /* create the thread for timer and interrupt handling */
    if (pthread_create(&thread, NULL, timing, (void *) NULL)) {
        printf("can't create timing thread\n");
        exit(1);
    }
    //pthread_setschedprio(thread, sched_get_priority_max(SCHED_RR));
    printf("\n");
}

/*
 *	This function is to stop the I/O devices. It is
 *	called from the CPU simulation on exit.
 */
void exit_io(void)
{

    /* close line printer files */
    if (lpt1 != 0) close(lpt1);
    if (lpt2 != 0) close(lpt2);

//    fp_stop();

    /* close network connections */
/*	for (i = 0; i < NUMNSOC; i++) {
		if (ncons[i].ssc)
			close(ncons[i].ssc);
	}
*/

    /* stop 10ms interrupt timer */
}

/*
 *	This is the main handler for all IN op-codes,
 *	called by the simulator. It calls the input
 *	function for port addr.
 */
BYTE io_in(BYTE addrl, BYTE addrh)
{
//	extern void wait_step(void);

        UNREFERENCED_PARAMETER(addrh);

	io_port = addrl;
	io_data = (*port_in[addrl]) ();
//if (io_port!=0x33) printf("input %02x from port %02x\r\n", io_data, io_port);

//	cpu_bus = CPU_WO | CPU_INP;

//	fp_clock += 3;
//	fp_led_address = (addrh << 8) + addrl;
//	fp_led_data = io_data;
//	fp_sampleData();
//	wait_step();

	return io_data;
}

/*
 *	This is the main handler for all OUT op-codes,
 *	called by the simulator. It calls the output
 *	function for port addr.
 */
void io_out(BYTE addrl, BYTE addrh, BYTE data)
{
//	extern void wait_step();

        UNREFERENCED_PARAMETER(addrh);

	io_port = addrl;
	io_data = data;
	(*port_out[addrl]) (data);
//if (io_port!=0x33) printf("output %02x to port %02x\r\n", io_data, io_port);

//	cpu_bus = CPU_OUT;

//	fp_clock += 6;
//	fp_led_address = (addrh << 8) + addrl;
//	fp_led_data = 0xff;
//	fp_sampleData();
//	wait_step();
}

static void wiring_tick(void) { // 10mS ts
    int i;
    static int count=0;

    if (++count<50) return;

    // 500 mS ts
    count=0;
    for (i=0; i<4; i++)
        fp_wrInt(80+i,c16fdc_rdInt(i));
}


/*
 *	Thread for timing and interrupts
 */
// this is based on Tstates. 1mS at 4MHz.
// It runs faster or slower with the system, smoother debugging
void ts1ms(void) {
//    static int count=-5000; // waits 5sec before generating 10mS interrupts
    static int count=-5; // waits 5 msec before generating 10mS interrupts

    /* reset disk index pulse */
//9    index_pulse = 0;

    /* check for interrupts from highest priority to lowest */
    fdc_uart_tick();

    if (++count<0) return;

    if ((count %   10) == 0) {    // 10mS
        c16fdc_tick();
        wiring_tick();
    }
    if ((count % 1000) == 0) vt100_tick();  //  1S
    if (count == 1000) count= 0;
}

void *timing(void *arg)
{
    UNREFERENCED_PARAMETER(arg);

    while (1) {	/* 2 msec per loop iteration (1mS inaccurate on Win)*/

        /* reset disk index pulse */
//        index_pulse = 0;

        /* check for interrupts from highest priority to lowest */
//        fdc_uart_tick();

        /* if last interrupt not acknowledged by CPU no new one yet */
        if (int_int)
            goto next;

//		/* EOJ from disk */ // (allow FDC to generate "RST 38"s if EOJ)
//		if ((fdc_flags & 1) && (uart0a_int_mask & 4)) {
//			uart0a_int = 0xd7;
//			uart0a_int_pending = 1;
//			int_data = 0xd7;
//			int_int = 1;
//			//printf("UART 0a EOJ interrupt\r\n");
//			goto next;
//		}

        fdc_uart_rt10ms();
//        if (++count >= 5) {
//            interrupt(1);
//            count=0;
//        }

next:
        os_sleep(10);  // 2 mS, more or less;
    }

    /* never reached, this thread is running endless */
    pthread_exit(NULL);
    return NULL;
}


/*
   used for hibernation
   - slow the CPU (and load on host system) if nothing is happening

   On a mobile phone app, this reduces battery consumption and extends life.
   On a desktop this stops host CPU sitting at 100% utilisation.
*/
static int activityFlag= 0;

void setActivityFlag(void) {
    activityFlag= 1;
}

int getActivityFlag(void) {    // test clear flag
    int x;

    x= activityFlag;
    activityFlag= 0;
    return x;
}
