/*
 * G2FDC - Emulation of a GEN2 Floppy Disk Controller
 * Copyright (C) 2018 Greg Sydney-Smith
 *
 * History:
 * 23-SEP-2018 first version. Based on earlier work by Udo Munk.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "sim.h"
#include "simglb.h"
#include "g2fdc.h"
#include "memory.h"

#ifdef _WIN32
#define OPENRO O_RDONLY|O_BINARY
#define OPENRW O_RDWR|O_BINARY
#else
#define OPENRO O_RDONLY
#define OPENRW O_RDWR
#endif

#define DISKS 4



/*
 * disk definitions 5.25"/8" drives, single/double density,
 * single/double sided
 */
enum Disk_mode    { READWRITE, READONLY };

typedef struct {
	char *fn;			/* filename of disk image */
	int tracks;			/* # of tracks */
	int spt0;			/* # sectors on track 0 */
	int sectors;			/* # sectors on tracks > 0 */
	int type;
	enum Disk_mode disk_m;		/* R/W or R/O mode */
} Diskdef;

static int dma;
static int disk = 0;		/* current selected disk # */
static int fdc_track = 0;	/* FDC track register */
static BYTE fdc_sec = 1;	/* FDC sector register */
static BYTE fdc_stat;		/* FDC status register */
static char fn[4096];		/* path/filename for disk image */
//

/* these are our disk drives, 8" DSDD initially */
static Diskdef disks[DISKS] = {
    { "drivea.dsk",  154, 64, 64, 21, READWRITE },
    { "driveb.dsk",  154, 64, 64, 21, READWRITE },
    { "drivec.dsk",  154, 64, 64, 21, READWRITE },
    { "drived.dsk",  154, 64, 64, 21, READWRITE },
};

/*
dbg:
 0 off
>0 commands (& their completion for those that set BUSY)
>1 OUTs to significant ports (eg track, sector; not data stream out)
>2 polling (INs of status ports, where the status changes
*/
static int dbg=0;

int g2fdc_config(char *name, char *val) {
    int i;
    char *names[] = { "debug", NULL };

    if (val==NULL) val="0";
    for (i=0; names[i]!=NULL && stricmp(name,names[i])!=0; i++) ;
    switch (i) {
    case 0: dbg    = atoi(val); break;
    }
    return 0;
}

int g2fdc_init(void) {
    return 0;
}


/*
 * find and set path for disk images
 */
void dsk_path(void) {
    struct stat sbuf;

    /* if option -d is used disks are there */
    if (diskdir != NULL) {
        strcpy(fn, diskd);
    } else {
        /* if not first try ./disks */
        if ((stat("./disks", &sbuf) == 0) && S_ISDIR(sbuf.st_mode)) {
            strcpy(fn, "./disks");
        /* nope, then DISKSDIR as set in Makefile */
        } else {
            strcpy(fn, DISKSDIR);
        }
    }
}

/*
 * configure drive and disk geometry from image file size
 * type= 8SS?16:8DS?20:8HD?24:5SD?32:5DS?36:252+SD?0:DD?1?DD0?2
 */
void config_disk(int fd)
{
    struct stat s;

    fstat(fd, &s);
    disks[disk].disk_m = (s.st_mode & S_IWUSR) ? READWRITE : READONLY;

//printf("size=%ld\n",s.st_size);
    switch (s.st_size) {
    case 77*26*128:		/* 8SSSD for backward compatability */
        disks[disk].tracks  =   77;
        disks[disk].spt0    =   26;
        disks[disk].sectors =   26;
        disks[disk].type    =   16;
        break;

    case 77*16*512:		/* 8SSDD */
        disks[disk].tracks  =   77;
        disks[disk].spt0    =   26;
        disks[disk].sectors =   64;
        disks[disk].type    =   17;
        break;

    case 26*128+76*16*512:	/* 8SSDD0 */
        disks[disk].tracks  =   77;
        disks[disk].spt0    =   26;
        disks[disk].sectors =   64;
        disks[disk].type    =   18;
        break;

    case 154*26*128:		/* 8DSSD */
        disks[disk].tracks  =  154;
        disks[disk].spt0    =   26;
        disks[disk].sectors =   26;
        disks[disk].type    =   20;
        break;

    case 154*16*4*128:		/* full 8" DSDD */
        disks[disk].tracks  =  154;
        disks[disk].spt0    =   64;
        disks[disk].sectors =   64;
        disks[disk].type    =   21;
        break;

    case 26+128+153*16*512:	/* 8DSDD0 */
        disks[disk].tracks  =  154;
        disks[disk].spt0    =   26;
        disks[disk].sectors =   64;
        disks[disk].type    =   22;
        break;

    case 8192*1024:		/* 8MB HDD */
        disks[disk].tracks  = 1024;
        disks[disk].spt0    =   64;
        disks[disk].sectors =   64;
        disks[disk].type    =   25;
        break;

    case 40*18*128:		/* 5SSSD for backward compatability */
        disks[disk].tracks  =   40;
        disks[disk].spt0    =   18;
        disks[disk].sectors =   18;
        disks[disk].type    =   32;
        break;

    case 40*10*512:		/* 5SSDD */
        disks[disk].tracks  =   40;
        disks[disk].spt0    =   40;
        disks[disk].sectors =   40;
        disks[disk].type    =   33;
        break;

    case 18*128+39*10*512:	/* 5SSDD0 */
        disks[disk].tracks  =   40;
        disks[disk].spt0    =   18;
        disks[disk].sectors =   40;
        disks[disk].type    =   34;
        break;

    case 80*18*128:		/* 5DSSD */
        disks[disk].tracks  =   80;
        disks[disk].spt0    =   18;
        disks[disk].sectors =   18;
        disks[disk].type    =   36;
        break;

    case 80*10*512:		/* 5DSDD */
        disks[disk].tracks  =   80;
        disks[disk].spt0    =   40;
        disks[disk].sectors =   40;
        disks[disk].type    =   37;
        break;

    case 18*128+79*10*512:	/* 5DSDD0 */
        disks[disk].tracks  =   80;
        disks[disk].spt0    =   18;
        disks[disk].sectors =   40;
        disks[disk].type    =   38;
        break;

    default:
        disks[disk].tracks  =   -1;
        disks[disk].type    =  252;
        break;
    }
}

/*
 * calculate disk image seek position for track/sector/side
 */
long get_pos(void)
{
    long pos;

    pos = (fdc_sec - 1);
    if (fdc_track > 0) pos += disks[disk].spt0;
    if (fdc_track > 1) pos += (fdc_track-1) * disks[disk].sectors;
    pos *= 128;
    return pos;
}


void g2fdc_disk_out(BYTE data) {
    disk = data;
}

void g2fdc_dmahi_out(BYTE data) {
    dma = (data*256) + (dma & 0x00ff);
}

void g2fdc_dmalo_out(BYTE data) {
    dma = (dma & 0xff00) + data;
}

void g2fdc_tkhi_out(BYTE data) {
    fdc_track = (data*256) + (fdc_track & 0x00ff);
}

void g2fdc_tklo_out(BYTE data) {
    fdc_track = (fdc_track & 0xff00) + data;
}

void g2fdc_sec_out(BYTE data) {
    fdc_sec = data;
}





static void cmd_writeTrk(void) {
    BYTE *p;
    long pos;
    int fd, i, secsiz;

    if (dbg>0)
        fprintf(stderr,"G2FDC: writeTrk Drv %c Tk %2d\n",(char)('A'+disk),fdc_track);

    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if (fdc_track == 0) {
        unlink(fn);
        if ((fd = open(fn, OPENRW|O_CREAT, 0644)) == -1)
            { fdc_stat = 0x80; return; }
        close(fd);
        disks[disk].tracks = 0;
        disks[disk].spt0   = 0;
        disks[disk].sectors= 0;
    }

    if ((fd = open(fn, OPENRW)) == -1)
        { fdc_stat = 0x40; return; }

    fdc_sec= 1;
    for (fdc_stat=0, i=0, p=memory+dma; 1; ) {

        /* find sector address mark
           (longest tk leadin and between sector seq is <200 bytes) */
        for (i=0; i<200 && p[i]!=0xfe; i++) ;
        if (i==200) break;

        p+=i; p++;                           /* to track,side,sector,size */
        secsiz= (p[3]==2)? 0x0200 : 0x0080;  /* DD / SD */

        /* find sector data mark */
        for (i=0; i<200 && p[i]!=0xfb; i++) ;
        if (i==200) break;

        p+=i; p++; /* past data mark */

        /* seek to sector */
        pos = get_pos();
        if (dbg>1)
            fprintf(stderr,"G2FDC: writeTrk. fdc_track=%d fdc_sec=%d pos=%08lx\n",
                fdc_track,fdc_sec,pos);
        if (lseek(fd, pos, SEEK_SET) == -1L)
            { fdc_stat= 0x10; close(fd); return; }

        /* write sector */
        fdc_stat = (write(fd,p,secsiz) == secsiz)? 0 : 0x10;
        if (fdc_stat) break;

        /* move past sector */
        p+=secsiz;

        /* check we have a CRC mark at the end */
        if (p[0]!=0xf7) break;

        fdc_sec += (secsiz==0x200)? 4 : 1;
    }

    if (fdc_track >= disks[disk].tracks)    /* 0-76 = 77 tracks */
        disks[disk].tracks = fdc_track+1;
    if (fdc_sec   >  disks[disk].spt0    && fdc_track==0)   /* 1-26 = 26 sectors */
        disks[disk].spt0   = fdc_sec-1;
    if (fdc_sec   >  disks[disk].sectors && fdc_track==1)
        disks[disk].sectors= fdc_sec-1;

    if (dbg>0)
        fprintf(stderr,"G2FDC: writeTrk complete. %d sectors written.\n",
            fdc_stat? 0 : fdc_sec-1);
    close(fd);
}


static void cmd_write128(void) {
    long pos;
    int lastsec, dcnt, fd;
    BYTE buf[128];		/* buffer for one sector */

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRW)) == -1)
        { fdc_stat= 0x80; return; }
 
    /* get drive and disk geometry */
    config_disk(fd);
    if (disks[disk].tracks < 1)
        { fdc_stat = 0x10; close(fd); return; }
 
    /* check track/sector */
    lastsec = disks[disk].sectors;
    if ((fdc_track >= disks[disk].tracks) || (fdc_sec == 0) || (fdc_sec > lastsec))
        { fdc_stat= 0x10; close(fd); return; }

    /* seek to sector */
    pos = get_pos();
    if (lseek(fd, pos, SEEK_SET) == -1L)
        { fdc_stat= 0x10; close(fd); return; }

    /* write data bytes into the sector buffer */
    for (dcnt=0; dcnt<128; dcnt++)
//        buf[dcnt] = dma_read(dma+dcnt);
        buf[dcnt] = memory[dma+dcnt];

    fdc_stat = (write(fd, buf, 128) == 128)? 0 : 0x10;	/* sector not found */
    close(fd);
}

static void cmd_read128(void) {
    long pos;
    int lastsec, dcnt, fd;
    BYTE buf[128];		/* buffer for one sector */

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRO)) == -1)
        { fdc_stat= 0x80; printf("G2FDC: disk %s not found\n",fn); return; }

    /* get drive and disk geometry */
    config_disk(fd);
    if (disks[disk].tracks < 0)
        { fdc_stat= 0x10; printf("G2FDC: unknown disk type\n"); close(fd); return; }

    /* check track/sector */
    lastsec = disks[disk].sectors;
    if ((fdc_track >= disks[disk].tracks) || (fdc_sec<1) || (fdc_sec>lastsec)) {
        printf("G2FDC: track %d sector %d seek error\n",fdc_track,fdc_sec);
        close(fd); fdc_stat= 0x10; return;
    }

    /* seek to sector */
    pos = get_pos();
    if (lseek(fd, pos, SEEK_SET) == -1L)
        { fdc_stat= 0x10; printf("G2FDC: seek failed\n"); close(fd); return; }

    if (dbg>0)
        printf("G2FDC RD. disk=%d, track=%4d, sec=%2d, pos=%08lX, dma=%04x\n",
            disk,fdc_track,fdc_sec,pos,dma);

    /* read the sector */
    if (read(fd, &buf[0], 128) != 128)
        { fdc_stat= 0x10; printf("G2FDC: read error\n"); close(fd); return; }

    for (dcnt=0; dcnt<128; dcnt++)
//        dma_write(dma+dcnt,buf[dcnt]);
        memwrt(dma+dcnt,buf[dcnt]);

    close(fd);
    fdc_stat = 0;
}

static void cmd_readTrk(void) {
    int lastsec;

    lastsec= (fdc_track==0)? disks[disk].spt0 : disks[disk].sectors;
    for (; fdc_sec<=lastsec; fdc_sec++) {
        cmd_read128();
        dma+=128;
    }
    fdc_sec=1;
    fdc_track++;
}

// fdc_stat is set by disksize (5"/8" SS/DS SD/DD/DD0 HD).
// if set appropriately, BIOS can use this to auto recognise disk type.
static void cmd_readType1(void) {
    int fd;

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRO)) == -1)
        { fdc_stat= 0x01; return; } // printf("G2FDC: disk %s not found\n",fn); return; }

    /* get drive and disk geometry */
    config_disk(fd);
    fdc_stat= disks[disk].type;

    close(fd);
}

// fdc_stat is set to byte 127 of boot sector.
// if set appropriately, BIOS can use this to auto recognise disk type.
// never set it to <' ' (reserved for error codes).
static void cmd_readType2(void) {
    int fd;
    BYTE buf[128];		/* buffer for one sector */

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRO)) == -1)
        { fdc_stat= 0x01; return; } // printf("G2FDC: disk %s not found\n",fn); return; }

    /* seek to sector */
    if (lseek(fd, 0, SEEK_SET) == -1L)
        { close(fd); fdc_stat= 0x02; return; } // printf("G2FDC: seek error\n"); return; }

    /* read the sector */
    if (read(fd, &buf[0], 128) != 128)
        { close(fd); fdc_stat= 0x03; return; } // printf("G2FDC: read error\n"); return; }

    fdc_stat= buf[127];
    close(fd);
}

void g2fdc_cmd_out(BYTE data) {
    switch (data) {
    case 0: cmd_write128(); break;
    case 1: cmd_read128();  break;
    case 2: cmd_writeTrk(); break;
    case 3: cmd_readTrk();  break;
    case 5: cmd_readType1(); break;
    case 7: cmd_readType2(); break;
    default: printf("G2FDC: unsupported cmd\n");
    }
}



BYTE g2fdc_status_in(void) {
    return fdc_stat;
}

void g2fdc_tick(void) { // every 10mS
}

/*
 * Reset G2FDC, placeholder for now, consult manuals
 */
void g2fdc_reset(void) {
}
