/*
 * Z80SIM  -  a Z80-CPU simulator
 *
 * Common I/O devices used by various simulated machines
 *
 * Copyright (C) 2014-2017 by Udo Munk
 * Copyright (C) 2017 changes by Greg Sydney-Smith
 *
 * Emulation of a Cromemco 4FDC/16FDC S100 board
 *
 * History:
 * 20-DEC-2014 first version
 * 28-DEC-2014 second version with 16FDC, CP/M 2.2 boots
 * 01-JAN-2015 fixed 16FDC, machine now also boots CDOS 2.58 from 8" and 5.25"
 * 01-JAN-2015 fixed frontpanel switch settings, added boot flag to fp switch
 * 12-JAN-2015 implemented dummy write track, so that programs won't hang
 * 22-JAN-2015 fixed buggy ID field, fake index pulses for 300/360 RPM
 * 26-JAN-2015 implemented write track to format SS/SD disks properly
 * 02-FEB-2015 implemented DS/DD disk formats
 * 05-FEB-2015 implemented DS/SD disk formats
 * 06-FEB-2015 implemented write track for all formats
 * 12-FEB-2015 implemented motor control, so that a 16FDC is recogniced by CDOS
 * 20-FEB-2015 bug fixes for 1.25 release
 * 08-MAR-2016 support user path for disk images
 * 13-MAY-2016 find disk images at -d <path>, ./disks and DISKDIR
 * 22-JUL-2016 added support for read only disks
 * 22-JUN-2017 added reset function
 * 26-JUL-2017 fixed buggy index pulse implementation
 * 14-NOV-2017 gss. Mods for Win32.
 * 26-NOV-2017 gss. Don't refuse Cmds whilst BUSY. And dbg changes, fdc_tick()
 * 19-AUG-2018 gss. Add 2MB disk size (8" 128Tk 128SPT 128b/sec)
 * 16-SEP-2018 gss. Rm 2MB disk. Now have DSDD CP/M BIOS going instead.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "sim.h"
#include "simglb.h"
#include "gss-fdc.h"
#include "memory.h"

#ifdef _WIN32
#define OPENRO O_RDONLY|O_BINARY
#define OPENRW O_RDWR|O_BINARY
#else
#define OPENRO O_RDONLY
#define OPENRW O_RDWR
#endif

#define DISKS 5



/*
 * disk definitions 5.25"/8" drives, single/double density,
 * single/double sided
 */
enum Disk_mode    { READWRITE, READONLY };

typedef struct {
	char *fn;			/* filename of disk image */
	int tracks;			/* # of tracks */
	int sectors;			/* # sectors on tracks > 0 side 0 */
	enum Disk_mode disk_m;		/* R/W or R/O mode */
} Diskdef;

static int dma;
static int disk = 0;		/* current selected disk # */
static BYTE fdc_track = 0;	/* FDC track register */
static BYTE fdc_sec = 1;	/* FDC sector register */
static BYTE fdc_stat;		/* FDC status register */
static char fn[4096];		/* path/filename for disk image */
//

/* these are our disk drives, 8" SS SD initially */
static Diskdef disks[DISKS] = {
    { "drivea.dsk",  154, 64, READWRITE },
    { "driveb.dsk",  154, 64, READWRITE },
    { "drivec.dsk",  154, 64, READWRITE },
    { "drived.dsk",  154, 64, READWRITE },
    { "drivee.dsk", 1024, 64, READWRITE }
};

/*
16fdc switches & defaults:
sw1 OFF RDOS-II is not defeated (ROM is present. not used here.)
sw2 ON  RDOS-II disabled after boot (allows "out (40),a" to disable ROM. we're RAM - not used)
sw3 ON  operating system loads automatically  (autoboot)
sw4 OFF operating system can format diskettes
sw5 OFF baud rate not pre-set (probably HW)
sw6 OFF unused
sw7 OFF \ RDOS 2.52 boot drive: 00=A, 01=B 10=C 11=D
sw8 OFF /

dbg:
 0 off
>0 commands (& their completion for those that set BUSY)
>1 OUTs to significant ports (eg track, sector; not data stream out)
>2 polling (INs of status ports, where the status changes
*/
static int dbg=0;
static int switches[1+8]= { 99,0,1,1,0,0,0,0,0 }; // 0xf9; // 64=autoboot 16

int gss_fdc_config(char *name, char *val) {
    int i;
    char *names[] = { "debug", "sw1", "sw2", "sw3", "sw4", "sw5", "sw6", "sw7", "sw8", NULL };

    if (val==NULL) val="0";
    for (i=0; names[i]!=NULL && stricmp(name,names[i])!=0; i++) ;
    switch (i) {
    case 0: dbg    = atoi(val); break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8: switches[i]= atoi(val); break;
    }
    return 0;
}

int gss_fdc_init(void) {
    return 0;
}


/*
 * find and set path for disk images
 */
void dsk_path(void) {
    struct stat sbuf;

    /* if option -d is used disks are there */
    if (diskdir != NULL) {
        strcpy(fn, diskd);
    } else {
        /* if not first try ./disks */
        if ((stat("./disks", &sbuf) == 0) && S_ISDIR(sbuf.st_mode)) {
            strcpy(fn, "./disks");
        /* nope, then DISKSDIR as set in Makefile */
        } else {
            strcpy(fn, DISKSDIR);
        }
    }
}

/*
 * configure drive and disk geometry from image file size
 */
void config_disk(int fd)
{
    struct stat s;

    fstat(fd, &s);
    disks[disk].disk_m = (s.st_mode & S_IWUSR) ? READWRITE : READONLY;

    switch (s.st_size) {
    case 154*16*4*128:		/* full 8" DS DD */
//    case 1256704:		/* 8" DS DD */
        disks[disk].tracks = 154;
        disks[disk].sectors = 64;
        break;

    case 8192*1024:		/* 8MB HDD */
        disks[disk].tracks  = 1024;
        disks[disk].sectors =   64;
        break;

    default:
        disks[disk].tracks = -1;
        break;
    }
}

/*
 * calculate disk image seek position for track/sector/side
 */
long get_pos(void)
{
    long pos;

    pos = ((fdc_track) * disks[disk].sectors + (fdc_sec - 1)) * 128;
    return pos;
}


void gss_fdc_disk_out(BYTE data) {
    disk = data;
}


void gss_fdc_dmahi_out(BYTE data) {
    dma = (data*256) + (dma & 0x00ff);
}

void gss_fdc_dmalo_out(BYTE data) {
    dma = (dma & 0xff00) + data;
}

void gss_fdc_tkhi_out(BYTE data) {
    fdc_track = (data & 0xff00) + (fdc_track & 0x00ff);
}

void gss_fdc_tklo_out(BYTE data) {
    fdc_track = (fdc_track & 0xff00) + (data & 0x00ff);
}

void gss_fdc_sec_out(BYTE data) {
    fdc_sec = data;
}





static void cmd_writeTrk(void) {
    BYTE *p;
    long pos;
    int fd, i, secsiz;

    if (dbg>0)
        fprintf(stderr,"gfdc: writeTrk Drv %c Tk %2d\n",(char)('A'+disk),fdc_track);

    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if (fdc_track == 0) {
        unlink(fn);
        if ((fd = open(fn, OPENRW|O_CREAT, 0644)) == -1)
            { fdc_stat = 0x80; return; }
        close(fd);
        disks[disk].tracks = 0;
        disks[disk].sectors= 0;
    }

    if ((fd = open(fn, OPENRW)) == -1)
        { fdc_stat = 0x40; return; }

    fdc_sec= 1;
    for (fdc_stat=0, i=0, p=memory+dma; 1; ) {

        /* find sector address mark
           (longest tk leadin and between sector seq is <200 bytes) */
        for (i=0; i<200 && p[i]!=0xfe; i++) ;
        if (i==200) break;

        p+=i; p++;                          /* to track,side,sector,size */
        secsiz= (p[3]==2)? 0x200 : 0x0080;  /* DD / SD */

        /* find sector data mark */
        for (i=0; i<200 && p[i]!=0xfb; i++) ;
        if (i==200) break;

        p+=i; p++; /* past data mark */

        /* seek to sector */
        pos = get_pos();
        if (dbg>1)
            fprintf(stderr,"gfdc: writeTrk. fdc_track=%d fdc_sec=%d pos=%08lx\n",
                fdc_track,fdc_sec,pos);
        if (lseek(fd, pos, SEEK_SET) == -1L)
            { fdc_stat= 0x10; close(fd); return; }

        /* write sector */
        fdc_stat = (write(fd,p,secsiz) == secsiz)? 0 : 0x10;
        if (fdc_stat) break;

        /* move past sector */
        p+=secsiz;

        /* check we have a CRC mark at the end */
        if (p[0]!=0xf7) break;

        fdc_sec += (secsiz==0x200)? 4 : 1;
    }

    if (fdc_track >= disks[disk].tracks)    /* 0-76 = 77 tracks */
        disks[disk].tracks = fdc_track+1;
    if (fdc_sec   >  disks[disk].sectors)   /* 1-26 = 26 sectors */
        disks[disk].sectors= fdc_sec-1;     /* spt same for all tks inc tk0 */

//printf("tracks=%d sectors=%d\n",disks[disk].tracks,disks[disk].sectors);
    if (dbg>0)
        fprintf(stderr,"gfdc: writeTrk complete. %d sectors written.\n",
            fdc_stat? 0 : disks[disk].sectors);
    close(fd);
}


static void cmd_write128(void) {
    long pos;
    int lastsec, dcnt, fd;
    BYTE buf[128];		/* buffer for one sector */

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRW)) == -1)
        { fdc_stat= 0x80; return; }
 
    /* get drive and disk geometry */
    config_disk(fd);
    if (disks[disk].tracks < 1)
        { fdc_stat = 0x10; close(fd); return; }
 
    /* check track/sector */
    lastsec = disks[disk].sectors;
    if ((fdc_track >= disks[disk].tracks) || (fdc_sec == 0) || (fdc_sec > lastsec))
        { fdc_stat= 0x10; close(fd); return; }

    /* seek to sector */
    pos = get_pos();
    if (lseek(fd, pos, SEEK_SET) == -1L)
        { fdc_stat= 0x10; close(fd); return; }

    /* write data bytes into the sector buffer */
    for (dcnt=0; dcnt<128; dcnt++)
//        buf[dcnt] = dma_read(dma+dcnt);
        buf[dcnt] = memory[dma+dcnt];

    fdc_stat = (write(fd, buf, 128) == 128)? 0 : 0x10;	/* sector not found */
    close(fd);
}

static void cmd_read128(void) {
    long pos;
    int lastsec, dcnt, fd;
    BYTE buf[128];		/* buffer for one sector */

    /* try to open disk image */
    dsk_path();
    strcat(fn, "/");
    strcat(fn, disks[disk].fn);
    if ((fd = open(fn, OPENRO)) == -1)
        { fdc_stat= 0x80; printf("FDC: disk %s not found\n",fn); return; }

    /* get drive and disk geometry */
    config_disk(fd);
    if (disks[disk].tracks < 0)
        { fdc_stat= 0x10; printf("FDC: unknown disk type\n"); close(fd); return; }

    /* check track/sector */
    lastsec = disks[disk].sectors;
    if ((fdc_track >= disks[disk].tracks) || (fdc_sec<1) || (fdc_sec>lastsec))
        { fdc_stat= 0x10; printf("FDC: track error\n"); close(fd); return; }

    /* seek to sector */
    pos = get_pos();
    if (lseek(fd, pos, SEEK_SET) == -1L)
        { fdc_stat= 0x10; printf("FDC: seek error\n"); close(fd); return; }

    if (dbg>0)
        printf("FDC RD. disk=%d, track=%4d, sec=%2d, pos=%08lX, dma=%04x\n",
            disk,fdc_track,fdc_sec,pos,dma);

    /* read the sector */
    if (read(fd, &buf[0], 128) != 128)
        { fdc_stat= 0x10; printf("FDC: read error\n"); close(fd); return; }

    for (dcnt=0; dcnt<128; dcnt++)
//        dma_write(dma+dcnt,buf[dcnt]);
        memwrt(dma+dcnt,buf[dcnt]);

    close(fd);
    fdc_stat = 0;
}

void gss_fdc_cmd_out(BYTE data) {
    switch (data) {
    case 0: cmd_write128(); break;
    case 1: cmd_read128();  break;
    case 2: cmd_writeTrk(); break;
    default: printf("Unsupported gss_fdc_cmd\n");
    }
}



BYTE gss_fdc_status_in(void) {
    return fdc_stat;
}

void gss_fdc_tick(void) { // every 10mS
}

/*
 * Reset FDC, placeholder for now, consult manuals
 */
void gss_fdc_reset(void) {
}
